@@ -542,6 +542,14 @@ read_conf(void)
           sen_default_encoding = sen_enc_sjis;
           break;
         }
+        if (*p == 'l' || *p == 'L') {
+          sen_default_encoding = sen_enc_latin1;
+          break;
+        }
+        if (*p == 'k' || *p == 'K') {
+          sen_default_encoding = sen_enc_koi8r;
+          break;
+        }
         if (*p == 'n' || *p == 'N') {
           sen_default_encoding = sen_enc_none;
           break;
@@ -600,7 +608,7 @@ build_flags(int flags)
     case SEN_INDEX_MORPH_ANALYSE :
       return flags | SEN_SYM_WITH_SIS;
     case SEN_INDEX_NGRAM :
-      return flags | SEN_SYM_WITH_SIS;
+      return flags & ~SEN_SYM_WITH_SIS;
     case SEN_INDEX_DELIMITED :
       return flags & ~SEN_SYM_WITH_SIS;
     default :
@@ -643,7 +651,6 @@ sen_index_create(const char *path, int key_size,
 	  i->store = NULL;
 	}
 	if (!(flags & SEN_INDEX_WITH_STORE) || i->store) {
-	  // pthread_mutex_init(&i->write_mutex, NULL);
 	  sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
 	  return i;
 	}
@@ -686,7 +693,6 @@ sen_index_open(const char *path)
 	  i->store = NULL;
 	}
 	if(!(i->lexicon->flags & SEN_INDEX_WITH_STORE) || i->store) {
-	  // pthread_mutex_init(&i->write_mutex, NULL);
 	  sen_log("index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
 	  return i;
 	}
@@ -732,7 +738,6 @@ sen_index_create_with_keys(const char *path, sen_sym *keys,
 	i->store = NULL;
       }
       if (!(flags & SEN_INDEX_WITH_STORE) || i->store) {
-	// pthread_mutex_init(&i->write_mutex, NULL);
 	sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
 	return i;
       }
@@ -771,7 +776,6 @@ sen_index_open_with_keys(const char *path, sen_sym *keys)
 	i->store = NULL;
       }
       if(!(i->lexicon->flags & SEN_INDEX_WITH_STORE) || i->store) {
-	// pthread_mutex_init(&i->write_mutex, NULL);
 	sen_log("index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
 	return i;
       }
@@ -787,7 +791,6 @@ sen_rc
 sen_index_close(sen_index *i)
 {
   if (!i) { return sen_invalid_argument; }
-  // pthread_mutex_destroy(&i->write_mutex);
   if (!i->foreign_keyp) { sen_sym_close(i->keys); }
   sen_sym_close(i->lexicon);
   sen_inv_close(i->inv);
@@ -885,6 +888,39 @@ sen_index_info(sen_index *i, int *key_size, int *flags,
   return sen_success;
 }
 
+sen_rc
+sen_index_lock(sen_index *i, int timeout)
+{
+  uint32_t count;
+  for (count = 0;; count++) {
+    uint32_t lock;
+    SEN_ATOMIC_ADD_EX(&i->keys->header->lock, 1, lock);
+    if (lock) {
+      SEN_ATOMIC_ADD_EX(&i->keys->header->lock, -1, lock);
+      if (!timeout || (timeout > 0 && timeout == count)) { break; }
+      usleep(1);
+      continue;
+    }
+    return sen_success;
+  }
+  return sen_other_error;
+}
+
+sen_rc
+sen_index_unlock(sen_index *i)
+{
+  uint32_t lock;
+  SEN_ATOMIC_ADD_EX(&i->keys->header->lock, -1, lock);
+  return sen_success;
+}
+
+sen_rc
+sen_index_clear_lock(sen_index *i)
+{
+  i->keys->header->lock = 0;
+  return sen_success;
+}
+
 int
 sen_index_path(sen_index *i, char *pathbuf, int bufsize)
 {
@@ -911,11 +947,11 @@ index_add(sen_index *i, const void *key, const char *value)
   sen_inv_updspec **u;
   sen_id rid, tid, *tp;
   int len = strlen(value);
+  sen_rc r, rc = sen_success;
   sen_store_buf *sbuf = NULL;
-  // pthread_mutex_lock(&i->write_mutex);
   //  sen_log("add > (%x:%x)", i, key);
   if (!(rid = sen_sym_get(i->keys, key))) { return sen_invalid_argument; }
-  if (!(lex = sen_lex_open(i->lexicon, value, 1))) {
+  if (!(lex = sen_lex_open(i->lexicon, value, SEN_LEX_ADD|SEN_LEX_UPD))) {
     return sen_memory_exhausted;
   }
   if (i->store) { sbuf = sen_store_buf_open(len); }
@@ -936,15 +972,14 @@ index_add(sen_index *i, const void *key, const char *value)
     hint = sen_str_get_prefix_order(_sen_sym_key(i->lexicon, *tp));
     if (hint == -1) { hint = *tp; }
     // sen_log("inv_update > %d '%s'", *tp, _sen_sym_key(i->lexicon, *tp));
-    sen_inv_update(i->inv, *tp, *u, h, hint);
+    if ((r = sen_inv_update(i->inv, *tp, *u, h, hint))) { rc = r; }
     // sen_log("inv_update < %d '%s'", *tp, _sen_sym_key(i->lexicon, *tp));
     sen_inv_updspec_close(*u);
   }
-  // pthread_mutex_unlock(&i->write_mutex);
   sen_set_cursor_close(c);
   sen_set_close(h);
   //  sen_log("add < (%x:%d:%d) %d", key, rid, *((int *)key), strlen(value));
-  return sen_success;
+  return rc;
 }
 
 inline static sen_rc
@@ -961,7 +996,7 @@ index_del(sen_index *i, const void *key, const char *value)
     return sen_invalid_argument;
   }
   // sen_log("del > (%x:%d:%d) %d", key, rid, *((int *)key), strlen(value));
-  if (!(lex = sen_lex_open(i->lexicon, value, 0))) {
+  if (!(lex = sen_lex_open(i->lexicon, value, SEN_LEX_UPD))) {
     return sen_memory_exhausted;
   }
   h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
@@ -974,7 +1009,6 @@ index_del(sen_index *i, const void *key, const char *value)
   }
   sen_lex_close(lex);
   c = sen_set_cursor_open(h);
-  // pthread_mutex_lock(&i->write_mutex);
   while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
     if (*tp) {
       // sen_log("inv_delete > %d '%s'", *tp, _sen_sym_key(i->lexicon, *tp));
@@ -983,7 +1017,6 @@ index_del(sen_index *i, const void *key, const char *value)
     }
     sen_inv_updspec_close(*u);
   }
-  // pthread_mutex_unlock(&i->write_mutex);
   sen_set_cursor_close(c);
   sen_set_close(h);
   // sen_log("del < (%x)", key);
@@ -995,16 +1028,25 @@ sen_index_upd(sen_index *i, const void *key,
               const char *oldvalue, const char *newvalue)
 {
   // if (!strcmp(oldvalue, newvalue)) { return sen_success; }
-  sen_rc rc = sen_success;
+  sen_rc rc;
+  if ((rc = sen_index_lock(i, -1))) {
+    sen_log("index lock failed");
+    return rc;
+  }
   // sen_log("index_upd %p %x > '%s'->'%s'", i, key, oldvalue, newvalue);
-  if (oldvalue && *oldvalue) { rc = index_del(i, key, oldvalue); }
-  if (rc != sen_success) { return rc; }
+  if (oldvalue && *oldvalue) {
+    rc = index_del(i, key, oldvalue); 
+    if (rc != sen_success) { goto exit; }
+  }
   if (newvalue && *newvalue) {
     rc = index_add(i, key, newvalue);
   } else {
     rc = sen_sym_del(i->keys, key);
   }
   //  sen_log("index_upd %x < %d", key, rc);
+exit :
+  sen_index_unlock(i);
+  sen_inv_seg_expire(i->inv);
   return rc;
 }
 
@@ -1064,18 +1106,24 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   int j, hint;
   sen_value *v;
   sen_lex *lex;
+  sen_rc r, rc;
   sen_set_eh *eh;
   sen_set_cursor *c;
   sen_set *old, *new;
   sen_id rid, tid, *tp;
-  sen_rc rc = sen_success;
   sen_inv_updspec **u, **un;
-  // pthread_mutex_lock(&i->write_mutex);
+  if ((rc = sen_index_lock(i, -1))) {
+    sen_log("index lock failed");
+    return rc;
+  }
   if (newvalues) {
-    if (!(rid = sen_sym_get(i->keys, key))) { return sen_invalid_argument; }
+    if (!(rid = sen_sym_get(i->keys, key))) {
+      rc = sen_invalid_argument;
+      goto exit;
+    }
     new = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
     for (j = newvalues->n_values, v = newvalues->values; j; j--, v++) {
-      if ((lex = sen_lex_open(i->lexicon, v->str, 1))) {
+      if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_ADD|SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(new, &tid, (void **) &u);
@@ -1091,13 +1139,16 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
       new = NULL;
     }
   } else {
-    if (!(rid = sen_sym_at(i->keys, key))) { return sen_invalid_argument; }
+    if (!(rid = sen_sym_at(i->keys, key))) {
+      rc = sen_invalid_argument;
+      goto exit;
+    }
     new = NULL;
   }
   if (oldvalues) {
     old = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
     for (j = oldvalues->n_values, v = oldvalues->values; j; j--, v++) {
-      if ((lex = sen_lex_open(i->lexicon, v->str, 0))) {
+      if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(old, &tid, (void **) &u);
@@ -1132,7 +1183,7 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
     while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
       hint = sen_str_get_prefix_order(_sen_sym_key(i->lexicon, *tp));
       if (hint == -1) { hint = *tp; }
-      sen_inv_update(i->inv, *tp, *u, new, hint);
+      if ((r = sen_inv_update(i->inv, *tp, *u, new, hint))) { rc = r; }
       sen_inv_updspec_close(*u);
     }
     sen_set_cursor_close(c);
@@ -1140,7 +1191,8 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   } else {
     rc = sen_sym_del(i->keys, key);
   }
-  // pthread_mutex_unlock(&i->write_mutex);
+exit :
+  sen_index_unlock(i);
   return rc;
 }
 
@@ -1336,7 +1388,7 @@ sen_records_find(sen_records *r, const void *key)
   recinfo *ri;
   if (!r || !r->keys || r->record_unit != sen_rec_document) { return 0; }
   if (!(rid = sen_sym_at(r->keys, key))) { return 0; }
-  if (!(r->curr_rec = sen_set_at(r->records, &rid, (void **) &ri))) { return 0; }
+  if (!sen_set_at(r->records, &rid, (void **) &ri)) { return 0; }
   return ri->score;
 }
 
@@ -1877,7 +1929,8 @@ bt_pop(btr *bt)
 typedef enum {
   sen_wv_none = 0,
   sen_wv_static,
-  sen_wv_dynamic
+  sen_wv_dynamic,
+  sen_wv_constant
 } sen_wv_mode;
 
 inline static int
@@ -1891,6 +1944,8 @@ get_weight(sen_records *r, sen_id rid, int sid,
     return sid <= optarg->vector_size ? optarg->weight_vector[sid - 1] : 0;
   case sen_wv_dynamic :
     return optarg->func(r, _sen_sym_key(r->keys, rid), sid, optarg->func_arg);
+  case sen_wv_constant :
+    return optarg->vector_size;
   default :
     return 1;
   }
@@ -1944,8 +1999,8 @@ sen_index_similar_search(sen_index *i, const char *string,
     rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
     if (optarg->func) {
       wvm = sen_wv_dynamic;
-    } else if (optarg->weight_vector && optarg->vector_size) {
-      wvm = sen_wv_static;
+    } else if (optarg->vector_size) {
+      wvm = optarg->weight_vector ? sen_wv_static : sen_wv_constant;
     }
     for (j = 0, eh = sorted; j < limit; j++, eh++) {
       sen_set_element_info(h, eh, (void **) &tp, (void **) &w1);
@@ -2013,8 +2068,8 @@ sen_index_term_extract(sen_index *i, const char *string,
   policy = optarg->max_interval;
   if (optarg->func) {
     wvm = sen_wv_dynamic;
-  } else if (optarg->weight_vector && optarg->vector_size) {
-    wvm = sen_wv_static;
+  } else if (optarg->vector_size) {
+    wvm = optarg->weight_vector ? sen_wv_static : sen_wv_constant;
   }
   if (policy == TERM_EXTRACT_EACH_POST) {
     sen_records_reopen(r, sen_rec_section, sen_rec_none, 0);
@@ -2085,8 +2140,8 @@ sen_index_select(sen_index *i, const char *string,
     mode = optarg->mode;
     if (optarg->func) {
       wvm = sen_wv_dynamic;
-    } else if (optarg->weight_vector && optarg->vector_size) {
-      wvm = sen_wv_static;
+    } else if (optarg->vector_size) {
+      wvm = optarg->weight_vector ? sen_wv_static : sen_wv_constant;
     }
   }
   if (mode == sen_sel_similar) {
@@ -2157,7 +2212,7 @@ sen_index_select(sen_index *i, const char *string,
     break; \
   } \
 }
-	if (n == 1) {
+	if (n == 1 && !rep) {
 	  noccur = (*tis)->p->tf;
 	  tscore = (*tis)->p->score;
 	} else if (mode == sen_sel_near) {
@@ -2230,6 +2285,15 @@ exit :
   }
   sen_records_cursor_clear(r);
   bt_close(bt);
+#ifdef DEBUG
+  {
+    uint32_t segno = SEN_INV_MAX_SEGMENT, nnref = 0;
+    sen_io_mapinfo *info = i->inv->seg->maps;
+    for (; segno; segno--, info++) { if (info->nref) { nnref++; } }
+    sen_log("nnref=%d", nnref);
+  }
+#endif /* DEBUG */
+  sen_inv_seg_expire(i->inv);
   return rc;
 }
 
@@ -2324,7 +2388,7 @@ sen_records_heap_add(sen_records_heap *h, sen_records *r)
   if (h->n_entries >= h->n_bins) {
     int size = h->n_bins * 2;
     sen_records **bins = SEN_REALLOC(h->bins, sizeof(sen_records *) * size);
-    sen_log("expanded sen_records_heap to %d,%p", size, bins);
+    // sen_log("expanded sen_records_heap to %d,%p", size, bins);
     if (!bins) { return sen_memory_exhausted; }
     h->n_bins = size;
     h->bins = bins;
