@@ -549,7 +549,10 @@ sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults,
           }
         }
         if (cond->end_offset > snip_result->start_offset + snip->width) {
+          /* If a keyword gets across a snippet, */
+          /* it was skipped and never to be tagged. */
           cond->stopflag = 2;
+          sen_bm_TunedBM(cond, snip->nstr, nstr_blen, snip->encoding);
         } else {
           found_cond = 1;
           if (cond->count == 0) {
