@@ -574,7 +574,8 @@ normalize_utf8(sen_nstr *nstr)
     }
     for (; *p && p < pe; p += lp) {
       lp = sen_str_charlen((char *)p, sen_enc_utf8);
-      if (*p == ' ' && removeblankp) {
+      if ((*p == ' ' && removeblankp)
+	  || *p < 0x20 	/* skip unprintable ascii */ ) {
 	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
       } else {
 	memcpy(d, p, lp);
